function JaniaTable(table, pNormal, pEdit, pSelect, tBodyFactory) {
	this.table = table
	this.tBodyFactory = tBodyFactory ?
		tBodyFactory :
		function() {return document.createElement("TBODY")}
	this.tbody = this.tBodyFactory()
	this.table.appendChild(this.tbody)
	this.listeners = new Array()
	this._pNormal = pNormal
	this._pEdit = pEdit
	this._pSelect = pSelect
	this._selectedRows = new Array()
	this._eventMasks = null
	this._ms = new MultiSelection(0)
	this._ms.addListener(this)
	this.lens = null
}

// --------------------------------------------------------------------------
// Events
//
JaniaTable.prototype.addListener = function(l) {
	this.listeners.push(l)
}
JaniaTable.prototype.fireOnRowAdd = function(row, index) {
	for(var i = 0; i < this.listeners.length; i++) {
		this.listeners[i].onRowAdd(row, index)
	}
}
JaniaTable.prototype.fireOnSelect = function(row, index) {
	for(var i = 0; i < this.listeners.length; i++) {
		this.listeners[i].onSelect(row, index)
	}
}
JaniaTable.prototype.fireOnDeselect = function(row, index) {
	for(var i = 0; i < this.listeners.length; i++) {
		this.listeners[i].onDeselect(row, index)
	}
}
JaniaTable.prototype.fireOnEndEdit = function(row, oldData) {
	for(var i = 0; i < this.listeners.length; i++) {
		this.listeners[i].onEndEdit(row, oldData)
	}
}
JaniaTable.prototype.fireOnBeforeEndEdit = function(oldData, newData, row) {
	var cancel = false
	for(var i = 0; i < this.listeners.length; i++) {
		if(this.listeners[i].onBeforeEndEdit(oldData, newData, row)) {
			cancel = true
		}
	}
	return cancel
}

// --------------------------------------------------------------------------
// MultiSelection listener
//
JaniaTable.prototype.onPreClick = function(row, col, ctrl, shift) {
	return false
}
JaniaTable.prototype.onSelect = function(row, col) {
	this.selectAt(row)
}
JaniaTable.prototype.onDeselect = function(row) {
	this.unselectAt(row)
}
JaniaTable.prototype.onStartEdit = function(row, col) {
	this.startEditAt(row, col)
}
JaniaTable.prototype.onEndEdit = function(row) {
	return this.endEditAt(row)
}

// --------------------------------------------------------------------------
// Add and Remove
//
JaniaTable.prototype.clearAll = function() {
	this._selectedRows = new Array()
	this.table.removeChild(this.tbody)
	this.tbody = this.tBodyFactory()
	this.table.appendChild(this.tbody)
	this._ms.reset()
	this._ms.length = 0
}
JaniaTable.prototype.addRow = function(values, lockEdit) {
	return this.addRowAt(this.tbody.rows.length, values, lockEdit)
}
JaniaTable.prototype.addRowAt = function(index, values, lockEdit) {
	var newRow = this._getNormalPrototype(values).cloneNode(true)
	newRow._attachedData = values
	newRow._editLocked = lockEdit

	var emptyRow = this.tbody.insertRow(index)
	this.tbody.replaceChild(newRow, emptyRow)

	this._attachEventToCells(newRow)

	for(var i = 0; i < values.length && i < newRow.cells.length; i++) {
		if (values[i] && newRow.cells[i]) {
			newRow.cells[i].innerHTML = this._cutOffLongString(i, values[i])
			if(newRow.cells[i].applyInnerHtml) newRow.cells[i].applyInnerHtml()
		}
	}
	this._ms.length++
	if(lockEdit) this._ms.addEditLock(index)

	this.fireOnRowAdd(newRow, index)
	return newRow
}
JaniaTable.prototype._attachEventToCells = function(row) {
	var oThis = this
	for(var i = 0; i < row.cells.length; i++) {
		var cell = row.cells[i]
		if(this._eventMasks && this._eventMasks[i]) {
			var handler = this._eventMasks[i]
			this._attachEvent(cell.firstChild, "click", handler)
		} else {
			var firstChild = cell.firstChild ? cell.firstChild : ""

			if(firstChild.type == "checkbox") {
				var handler = function(e) {
					var src = e.srcElement ? e.srcElement : e.target
					var ctrl = true
					var shift = e.shiftKey
					oThis._ms.click(oThis.getRowIndex(row), src.cellIndex, ctrl, shift)
				}
				this._attachEvent(firstChild, "click", handler)
			} else if (firstChild.type == "text") {
				var handler = function(e) {
					if(e.keyCode != 13) return
					oThis._ms.clickAgain()
				}
				this._attachEvent(cell, "keydown", handler)
			} else {
				var handler = function(e) {
					var src = e.srcElement ? e.srcElement : e.target
					var ctrl = e.ctrlKey
					var shift = e.shiftKey
					oThis._ms.click(oThis.getRowIndex(row), src.cellIndex, ctrl, shift)
				}
				this._attachEvent(cell, "click", handler)
			}
		}
	}
}

JaniaTable.prototype.getRowAt = function(index) {
	return this.tbody.rows[index]
}
JaniaTable.prototype.removeRowAt = function(index) {
	var row = this.getRowAt(index)
	for(var i = 0; i < this._selectedRows.length; i++) {
		if(this._selectedRows[i] == row) {
			this._selectedRows.splice(i, 1)
			break;
		}
	}

	this.tbody.removeChild(row)
	this._ms.length--
	this._ms.reset()
}
JaniaTable.prototype.getRowIndex = function(row) {
	return row.rowIndex - this.tbody.rows[0].rowIndex
}
JaniaTable.prototype.getLength = function() {
	return this.tbody.rows.length
}

// --------------------------------------------------------------------------
// Sort
//
JaniaTable.prototype.setInitialSortStatus = function(colIndex, ascending) {
	this._sortColumnIndex = colIndex;
	this._ascending = ascending;
}
JaniaTable.prototype.sort = function(colIndex, ascending, beginIndex) {
	var columns = this._getColumns(colIndex)
	var temp = new Array()

	if(beginIndex) {
		for(var i = 0; i < beginIndex; i++) {
			temp.push(columns[i])
		}
		columns.splice(0, beginIndex)
	}
	columns.sort(JaniaTable._compare)
	if(!ascending) columns.reverse()

	if(beginIndex) {
		for(var i = 0; i < beginIndex; i++) {
			columns.splice(0, 0, temp[beginIndex - i - 1])
		}
	}

	this.table.removeChild(this.tbody)
	this.tbody = this.tBodyFactory()

	for(var i = 0; i < columns.length; i++) {
		this.tbody.appendChild(columns[i].row)
	}

	this.table.appendChild(this.tbody)

	this._sortColumnIndex = colIndex
	this._ascending = ascending
}
JaniaTable.prototype.toggleSort = function(colIndex, beginIndex) {
	if(this._sortColumnIndex == colIndex) {
		this.sort(colIndex, !this._ascending, beginIndex)
	} else {
		this.sort(colIndex, true, beginIndex)
	}
}

// --------------------------------------------------------------------------
// Edit mode
//
JaniaTable.prototype.setEdit = function(enabled) {
	this._ms.setEdit(enabled)
}
JaniaTable.prototype.startEditAt = function(row, col) {
	var from = this.getRowAt(row)
	var to = this._getEditPrototype(row._attachedData).cloneNode(true)
	this._replaceRow(from, "nodeValue", to, "value")
	try {
		to.cells[col].firstChild.focus();
		to.cells[col].firstChild.focus();
	} catch(e) {
		;
	}
}
JaniaTable.prototype.endEditAt = function(index) {
	var from = this.getRowAt(index)
	var oldData = from._attachedData
	var newData = new Array()
	for(var i = 0; i < oldData.length; i++) {
		newData.push(oldData[i])
	}
	
	for(var i = 0; i < from.cells.length; i++) {
		var node = from.cells[i].firstChild
		if(node.type != "checkbox") {
			newData.splice(i, 1, node.value)
		} else {
			newData.splice(i, 1, null)
		}
	}
	
	if(this.fireOnBeforeEndEdit(oldData, newData, this.getRowAt(index))) {
		for(var i = 0; i < from.cells.length; i++) {
			var node = from.cells[i].firstChild
			if(node.type != "checkbox") {
				node.value = oldData[i]
			}
		}
		return true
	}
	
	var to = this._getNormalPrototype(newData).cloneNode(true)
	this._replaceRow(from, "value", to, "nodeValue")
	for(var i = 0; i < to.cells.length; i++) {
		if(to.cells[i].syncInnerHtml) to.cells[i].syncInnerHtml()
		to._attachedData[i] = to.cells[i].innerHTML
		to.cells[i].innerHTML = this._cutOffLongString(i, to.cells[i].innerHTML)
	}

	this._selectedRows = new Array()

	this.fireOnEndEdit(to, oldData)
}

// --------------------------------------------------------------------------
// Selection
//
JaniaTable.prototype.setMultiSelection = function(enabled) {
	this._ms.setMultiSelection(enabled)
}
JaniaTable.prototype.select = function(row) {
	var from = row
	var to = this._getSelectPrototype(row._attachedData).cloneNode(true)
	to._selected = true
	this._selectedRows.push(to)
	this._replaceRow(from, "nodeValue", to, "nodeValue")
	this.fireOnSelect(to, this.getRowIndex(to))
}
JaniaTable.prototype.unselect = function(row) {
	var from = row
	var to = this._getNormalPrototype(row._attachedData).cloneNode(true)
	to._selected = false
	for(var i = 0; i < this._selectedRows.length; i++) {
		if(this._selectedRows[i] == from) {
			this._selectedRows.splice(i, 1)
			break
		}
	}
	this._replaceRow(from, "nodeValue", to, "nodeValue")
	this.fireOnDeselect(to, this.getRowIndex(to))
}
JaniaTable.prototype.selectAt = function(index) {
	this.select(this.getRowAt(index))
}
JaniaTable.prototype.unselectAt = function(index) {
	this.unselect(this.getRowAt(index))
}
JaniaTable.prototype.getSelectedRows = function() {
	return this._selectedRows
}

// --------------------------------------------------------------------------
// Etc
//
JaniaTable.prototype.setEventMasks = function(masks) {
	this._eventMasks = masks
}
JaniaTable.prototype.setMaxDisplayLength = function(lens) {
	this._lens = lens
}

// --------------------------------------------------------------------------
// Privates
//

JaniaTable.prototype._cutOffLongString = function(index, str) {
	if(!this._lens) return str
	if(this._lens[index] == 0) return str
	
	return Util.cutOffString(this._lens[index], str)
}
JaniaTable.prototype._getColumns = function(colIndex) {
	var rows = this.tbody.rows
	var length = rows.length
	var array = new Array(length)
	for (var i = 0; i < length; i++) {
		array[i] = {value: rows[i]._attachedData[colIndex], row: rows[i]}
	}
	return array;
}
JaniaTable.prototype._replaceRow = function(from, fromValue, to, toValue) {
	var froms = new Array()
	for(var i = 0; i < from.cells.length; i++) {
		froms.push(from.cells[i].firstChild)
	}
	
	this.tbody.replaceChild(to, from)
	
	for(var i = 0; i < to.cells.length; i++) {
		var fromNode = froms[i]
		if(!fromNode) {
			fromNode = document.createTextNode("")
		}
		if(!to.cells[i].firstChild) {
			var emptyNode = document.createTextNode("")
			to.cells[i].appendChild(emptyNode)
		}
		var toNode = to.cells[i].firstChild
		if(toNode.type != "checkbox") {
			if(fromValue == "nodeValue" && toValue == "value") {
				toNode[toValue] = from._attachedData[i]
			} else {
				try {
					toNode[toValue] = fromNode[fromValue]
				} catch (e) {
					toNode[toValue] = from._attachedData[i]
				}
			}
		}
	}
	
	var newAttachedData = new Array()
	for(var i = 0; i < from._attachedData.length; i++) {
		var temp = from._attachedData[i]
		newAttachedData.push(temp)
	}
	to._attachedData = newAttachedData

	this._attachEventToCells(to)
}
JaniaTable.prototype._attachEvent = function(obj, eventName, f) {
	if (obj.addEventListener) {
		obj.addEventListener(eventName, f, false)
	} else {
		obj.attachEvent("on" + eventName, f)
	}
}
JaniaTable._compare = function(n1, n2) {
	if (n1.value < n2.value) return -1
	if (n1.value > n2.value) return 1
	return 0
}
JaniaTable.prototype._getNormalPrototype = function(attachedData) {
	return this._pNormal
}
JaniaTable.prototype._getEditPrototype = function(attachedData) {
	return this._pEdit
}
JaniaTable.prototype._getSelectPrototype = function(attachedData) {
	return this._pSelect
}
